//* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <object.h>
#include <queue.h>
#include <packet.h>
#include <cmu-trace.h>
//#include <iostream.h>
#include "priq.h"
#include "mac-802_11e.h"


int Qmax;
double BB;
double realBB;
int    KK;

double t_e;
double Wq;
double idlesum;


FILE *file_handler = fopen("incoming.interval.txt", "w+");
int    num_pkt;
double prev_incoming_time;


typedef int (*PacketFilter)(Packet *, void *);

PriQ_List PriQ::prhead = { 0 };

static class PriQClass : public TclClass {
public:
  PriQClass() : TclClass("Queue/DTail/PriQ") {}
  TclObject* create(int, const char*const*) {
    return (new PriQ);
  }
} class_PriQ;


PriQ::PriQ() : DTail()
{
        bind("Prefer_Routing_Protocols", &Prefer_Routing_Protocols);
	bind("Max_Levels", &max_levels);
	bind("Levels", &levels);

// 	bind("MinBufferLimit_", &MinBufferLimit);

        bind ("Wq_", 	        &Wq);

	LIST_INSERT_HEAD(&prhead, this, link);
	flag = 0;
	cntl = 0;
	level = 0;
}

int
PriQ::command(int argc, const char*const* argv)
{
    Scheduler &s = Scheduler::instance();
    if (argc == 5 )
    { 
	if(strcmp(argv[1], "Prio") == 0){
	    if(!(atoi(argv[2]) > levels)){
		level = atoi(argv[2]);
		if(strcmp(argv[3], "PF") == 0){
		    pri_[level].setPF(atoi(argv[4]));
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "CW_MIN") == 0){
		    pri_[level].setCW_MIN(atoi(argv[4]));
// printf("PriQ::command.. Prio=%d, cwmin=%d \n", level, atoi(argv[4]));
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "CW_MAX") == 0){
		    pri_[level].setCW_MAX(atoi(argv[4]));
// printf("PriQ::command.. Prio=%d, cwmax=%d \n", level, atoi(argv[4]));
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "AIFS") == 0){
		    pri_[level].setAIFS(atoi(argv[4])); 
// printf("PriQ::command.. Prio=%d, aifs=%d \n", level, atoi(argv[4]));
		    return(TCL_OK);	      
		}
		if(strcmp(argv[3], "TXOPLimit") == 0){
// printf("PriQ::command.. Prio=%d, TXOPLimit=%lf \n", level, atof(argv[4]));
		    pri_[level].setTXOPLimit(atof(argv[4])); 
		    return(TCL_OK);	      
		}		
		if(strcmp(argv[3], "QueueSize") == 0){
		    pri_[level].setQueueSize(atoi(argv[4])); 
		    return(TCL_OK);	      
		}
		if(strcmp(argv[3], "IdleSum") == 0){
		    idlesum=atof(argv[4]); 
		    return(TCL_OK);	      
		}
	    }else return (TCL_ERROR);
	}
    }
    Tcl& tcl = Tcl::instance();
    if (argc == 4 )
    { 
	if(strcmp(argv[1], "getValue") == 0){
	    if(!(atoi(argv[2]) > levels)){
		level = atoi(argv[2]);
		if(strcmp(argv[3], "PF") == 0){
		    tcl.resultf("%d", pri_[level].getPF());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "CW_MIN") == 0){
		    tcl.resultf("%d", pri_[level].getCW_MIN());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "CW_MAX") == 0){
		    tcl.resultf("%d", pri_[level].getCW_MAX());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "AIFS") == 0){
		    tcl.resultf("%d", pri_[level].getAIFS());
		    return(TCL_OK);	      
		}
		if(strcmp(argv[3], "TXOPLimit") == 0){
		    tcl.resultf("%lf", pri_[level].getTXOPLimit());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "NumDrops") == 0){
		    tcl.resultf("%d", pri_[level].getNumDrops());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "NumNullQueueSize") == 0){
		    tcl.resultf("%d", pri_[level].getNumNullQueueSize());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "NumArrived") == 0){
		    tcl.resultf("%d", pri_[level].getNumArrived());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "SumInterArrvialTime") == 0){
		    tcl.resultf("%lf", pri_[level].getInterArrvialTime());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "NumAllOverflew") == 0){
		    tcl.resultf("%d", pri_[level].getNumAllOverflew());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "Qmax") == 0){
		    tcl.resultf("%d", Qmax);
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "QueueSizeInPkts") == 0){
		    tcl.resultf("%d", pri_[level].length());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "QueueSizeInBytes") == 0){
		    tcl.resultf("%d", pri_[level].byteLength());
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "MaxQueueSizeInPkts") == 0){
		    tcl.resultf("%d", pri_[level].limit());
// 		    tcl.resultf("%d", Qmax);
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "BB") == 0){
		    tcl.resultf("%lf", BB);
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "realBB") == 0){
		    tcl.resultf("%lf", realBB);
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "KK") == 0){
		    tcl.resultf("%d", KK);
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "GET_WQ") == 0){
		    tcl.resultf("%lf", tempWq);
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "IdleSum") == 0){
		    tcl.resultf("%lf", idlesum);
		    return(TCL_OK);
		}
		if(strcmp(argv[3], "MACServTime") == 0){
		    tcl.resultf("%lf", ((Mac802_11e*) target())->ServTime);
		    return(TCL_OK);
		}
	    }else return (TCL_ERROR);
	}
    }
    if (argc == 2 && strcasecmp(argv[1], "reset") == 0)
    {
	Terminate();
	//FALL-THROUGH to give parents a chance to reset
    }

    return DTail::command(argc, argv);
}

void
PriQ::recv(Packet *p, Handler *h)
{
    struct hdr_cmn *ch = HDR_CMN(p);

// printf("PriQ::recv.now=%lf..pkt %d \n", Scheduler::instance().clock(), ch->uid());

    if(flag == 0) {
	((Mac802_11e*) target())->queue_ = this;
	flag = 1;
	for(int i = 0; i < MAX_PRI; i++) {
	    pri_[i].setdrop(drop_);    
	}
    }
    
    if(Prefer_Routing_Protocols) {

                switch(ch->ptype()) {
                case PT_DSR:
                case PT_MESSAGE:
                case PT_TORA:
                case PT_AODV:
                        recvHighPriority(p, h);
                        break;

                default:
		    pri_recv(p, h);
                }
        }
        else {
	    pri_recv(p, h);
        }
}

void
PriQ::pri_recv(Packet *p, Handler *h)
{
    Scheduler &s = Scheduler::instance();
    level = PKT_LEVEL(p); 
    struct hdr_cmn *ch = HDR_CMN(p);
    /*  target_handle() is necessary to give the target to class Queue.  
     *  Otherwise the target is not known in class Queue 
     *  when dequeing packet :o(
     */
	struct hdr_mac802_11e *dh = HDR_MAC802_11E(p);
	int dst, src;
	dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);



// // // // //  Tianji Li, the ALT algorithm
// 	double curr_idletime;
// 	if ( level == 1  ) {
// 		t_a = Scheduler::instance().clock();
// 		if( pri_[level].getLen() <= 0)  {
// 			curr_idletime = t_a - ((Mac802_11e*) target())->t_e;
// 		} else {
// 			curr_idletime = 0.0;
// 		}
// 		if ( src == 16 ) { idlesum = idlesum + curr_idletime; }
// 	}
// // // // //  Tianji Li, the ALT algorithm




// // //  Tianji Li, the aBDP algorithm
   if ( ((Mac802_11e*) target())->addr() == 16 && ch->ptype() == PT_TCP && ch->size() != 40 ) {
 	if ( level == 1 ) {
		if ( num_pkt == 0 ) prev_incoming_time = Scheduler::instance().clock();
		num_pkt ++;
		double tempTime = Scheduler::instance().clock() - prev_incoming_time;
		prev_incoming_time = Scheduler::instance().clock();

		double m_out = ((Mac802_11e*) target())->ServTime;
		double t_target = ((Mac802_11e*) target())->Tdrain;

		Qmax = (int) ceil ( t_target / m_out ) + 40;

// printf("t_target=%lf, m_out=%lf, Qmax=%d \n", t_target, m_out, Qmax);

		if (   pri_[level].q_->length() > Qmax ) { 
				drop(p);
				if ( ((Mac802_11e*) target())->is_idle() && ((Mac802_11e*) target())->mhDefer_.busy()==0 && ((Mac802_11e*) target())->mhBackoff_.busy()==0 && ((Mac802_11e*) target())->mhRecv_.busy()==0 && ((Mac802_11e*) target())->mhSend_.busy()==0) {
					pri_[level].target_handle(target_);
					pri_[level].resume();
				}
				return;
		}
 	}
    }
// // // Tianji Li, the aBDP algorithm

    pri_[level].target_handle(target_);
    pri_[level].recv(p,h);
}    


void 
PriQ::recvHighPriority(Packet *p, Handler *)
  // insert packet at front of queue
{
    pri_[cntl].q_->enqueHead(p);
    pri_[cntl].target_handle(target_);
    if (pri_[cntl].q_->length() >= qlim_)
    {
      Packet *to_drop = pri_[cntl].q_->lookup(pri_[cntl].q_->length()-1);
      pri_[cntl].q_->remove(to_drop);
      drop(to_drop);
    }
  if (!pri_[cntl].blocked()) {
    /*
     * We're not blocked.  Get a packet and send it on.
     * We perform an extra check because the queue
     * might drop the packet even if it was
     * previously empty!  (e.g., RED can do this.)
     */
  

      pri_[cntl].recvHighPri(p);
      //target_->recv(p, &qh_); <- done in d-tail.cc
    }
} 
 
void 
PriQ::filter(PacketFilter filter, void * data)
  // apply filter to each packet in queue, 
  // - if filter returns 0 leave packet in queue
  // - if filter returns 1 remove packet from queue
{
    int i = 0;
    while (i < pri_[cntl].q_->length())
    {
      Packet *p = pri_[cntl].q_->lookup(i);
      if (filter(p,data))
	{
	  pri_[cntl].q_->remove(p); // decrements q len
	}
      else i++;
    }
}

Packet*
PriQ::filter(nsaddr_t id)
{
    Packet *p = 0;
    Packet *pp = 0;
    struct hdr_cmn *ch;
    
    for(p = pri_[cntl].q_->head(); p; p = p->next_) {
	ch = HDR_CMN(p);
	if(ch->next_hop() == id)
	    break;
	pp = p;
    }

	/*
	 * Deque Packet
	 */
     if(p) {
	if(pp == 0)
	    pri_[cntl].q_->remove(p);
	else
	    pri_[cntl].q_->remove(p, pp);
    }
    return p;
}



/*
 * Called at the end of the simulation to purge the IFQ.
 */
void
PriQ::Terminate()
{
    for(int i = 0; i< MAX_PRI; i++){
	Packet *p;
	while((p = pri_[i].deque())) {
	    drop(p, DROP_END_OF_SIMULATION);
	    //drop(p);
		
	}
    }
}

int PriQ::getLevels(){
    return levels;
}




